# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..core.serialization import convert_and_respect_annotation_metadata
from ..errors.bad_request_error import BadRequestError
from ..errors.conflict_error import ConflictError
from ..errors.forbidden_error import ForbiddenError
from ..errors.internal_server_error import InternalServerError
from ..errors.not_found_error import NotFoundError
from ..errors.payment_required_error import PaymentRequiredError
from ..errors.service_unavailable_error import ServiceUnavailableError
from ..errors.too_many_requests_error import TooManyRequestsError
from ..errors.unauthorized_error import UnauthorizedError
from ..types.not_found_error_body import NotFoundErrorBody
from ..types.number_from_string import NumberFromString
from ..types.permission_denied_error import PermissionDeniedError
from ..types.plan_limit_exceeded_error import PlanLimitExceededError
from ..types.rate_limit_error import RateLimitError
from ..types.unauthorized_error_body import UnauthorizedErrorBody
from .types.traces_create_otel_request_resource_spans_item import (
    TracesCreateOtelRequestResourceSpansItem,
)
from .types.traces_create_otel_response import TracesCreateOtelResponse
from .types.traces_create_request_resource_spans_item import (
    TracesCreateRequestResourceSpansItem,
)
from .types.traces_create_response import TracesCreateResponse
from .types.traces_get_analytics_summary_response import (
    TracesGetAnalyticsSummaryResponse,
)
from .types.traces_get_trace_detail_by_env_response import (
    TracesGetTraceDetailByEnvResponse,
)
from .types.traces_get_trace_detail_response import TracesGetTraceDetailResponse
from .types.traces_list_by_function_hash_response import (
    TracesListByFunctionHashResponse,
)
from .types.traces_search_by_env_request_attribute_filters_item import (
    TracesSearchByEnvRequestAttributeFiltersItem,
)
from .types.traces_search_by_env_request_sort_by import TracesSearchByEnvRequestSortBy
from .types.traces_search_by_env_request_sort_order import (
    TracesSearchByEnvRequestSortOrder,
)
from .types.traces_search_by_env_response import TracesSearchByEnvResponse
from .types.traces_search_request_attribute_filters_item import (
    TracesSearchRequestAttributeFiltersItem,
)
from .types.traces_search_request_sort_by import TracesSearchRequestSortBy
from .types.traces_search_request_sort_order import TracesSearchRequestSortOrder
from .types.traces_search_response import TracesSearchResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawTracesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def create(
        self,
        *,
        resource_spans: typing.Sequence[TracesCreateRequestResourceSpansItem],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[TracesCreateResponse]:
        """
        Parameters
        ----------
        resource_spans : typing.Sequence[TracesCreateRequestResourceSpansItem]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TracesCreateResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            "traces",
            method="POST",
            json={
                "resourceSpans": convert_and_respect_annotation_metadata(
                    object_=resource_spans,
                    annotation=typing.Sequence[TracesCreateRequestResourceSpansItem],
                    direction="write",
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesCreateResponse,
                    parse_obj_as(
                        type_=TracesCreateResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 402:
                raise PaymentRequiredError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PlanLimitExceededError,
                        parse_obj_as(
                            type_=PlanLimitExceededError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def createotel(
        self,
        *,
        resource_spans: typing.Sequence[TracesCreateOtelRequestResourceSpansItem],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[TracesCreateOtelResponse]:
        """
        Parameters
        ----------
        resource_spans : typing.Sequence[TracesCreateOtelRequestResourceSpansItem]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TracesCreateOtelResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/traces",
            method="POST",
            json={
                "resourceSpans": convert_and_respect_annotation_metadata(
                    object_=resource_spans,
                    annotation=typing.Sequence[
                        TracesCreateOtelRequestResourceSpansItem
                    ],
                    direction="write",
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesCreateOtelResponse,
                    parse_obj_as(
                        type_=TracesCreateOtelResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 402:
                raise PaymentRequiredError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PlanLimitExceededError,
                        parse_obj_as(
                            type_=PlanLimitExceededError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def search(
        self,
        *,
        start_time: str,
        end_time: str,
        query: typing.Optional[str] = OMIT,
        input_messages_query: typing.Optional[str] = OMIT,
        output_messages_query: typing.Optional[str] = OMIT,
        fuzzy_search: typing.Optional[bool] = OMIT,
        trace_id: typing.Optional[str] = OMIT,
        span_id: typing.Optional[str] = OMIT,
        model: typing.Optional[typing.Sequence[str]] = OMIT,
        provider: typing.Optional[typing.Sequence[str]] = OMIT,
        function_id: typing.Optional[str] = OMIT,
        function_name: typing.Optional[str] = OMIT,
        span_name_prefix: typing.Optional[str] = OMIT,
        has_error: typing.Optional[bool] = OMIT,
        min_tokens: typing.Optional[float] = OMIT,
        max_tokens: typing.Optional[float] = OMIT,
        min_duration: typing.Optional[float] = OMIT,
        max_duration: typing.Optional[float] = OMIT,
        attribute_filters: typing.Optional[
            typing.Sequence[TracesSearchRequestAttributeFiltersItem]
        ] = OMIT,
        limit: typing.Optional[float] = OMIT,
        offset: typing.Optional[float] = OMIT,
        sort_by: typing.Optional[TracesSearchRequestSortBy] = OMIT,
        sort_order: typing.Optional[TracesSearchRequestSortOrder] = OMIT,
        root_spans_only: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[TracesSearchResponse]:
        """
        Parameters
        ----------
        start_time : str

        end_time : str

        query : typing.Optional[str]

        input_messages_query : typing.Optional[str]

        output_messages_query : typing.Optional[str]

        fuzzy_search : typing.Optional[bool]

        trace_id : typing.Optional[str]

        span_id : typing.Optional[str]

        model : typing.Optional[typing.Sequence[str]]

        provider : typing.Optional[typing.Sequence[str]]

        function_id : typing.Optional[str]

        function_name : typing.Optional[str]

        span_name_prefix : typing.Optional[str]

        has_error : typing.Optional[bool]

        min_tokens : typing.Optional[float]

        max_tokens : typing.Optional[float]

        min_duration : typing.Optional[float]

        max_duration : typing.Optional[float]

        attribute_filters : typing.Optional[typing.Sequence[TracesSearchRequestAttributeFiltersItem]]

        limit : typing.Optional[float]

        offset : typing.Optional[float]

        sort_by : typing.Optional[TracesSearchRequestSortBy]

        sort_order : typing.Optional[TracesSearchRequestSortOrder]

        root_spans_only : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TracesSearchResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            "traces/search",
            method="POST",
            json={
                "startTime": start_time,
                "endTime": end_time,
                "query": query,
                "inputMessagesQuery": input_messages_query,
                "outputMessagesQuery": output_messages_query,
                "fuzzySearch": fuzzy_search,
                "traceId": trace_id,
                "spanId": span_id,
                "model": model,
                "provider": provider,
                "functionId": function_id,
                "functionName": function_name,
                "spanNamePrefix": span_name_prefix,
                "hasError": has_error,
                "minTokens": min_tokens,
                "maxTokens": max_tokens,
                "minDuration": min_duration,
                "maxDuration": max_duration,
                "attributeFilters": convert_and_respect_annotation_metadata(
                    object_=attribute_filters,
                    annotation=typing.Sequence[TracesSearchRequestAttributeFiltersItem],
                    direction="write",
                ),
                "limit": limit,
                "offset": offset,
                "sortBy": sort_by,
                "sortOrder": sort_order,
                "rootSpansOnly": root_spans_only,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesSearchResponse,
                    parse_obj_as(
                        type_=TracesSearchResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def gettracedetail(
        self, trace_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[TracesGetTraceDetailResponse]:
        """
        Parameters
        ----------
        trace_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TracesGetTraceDetailResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            f"traces/{jsonable_encoder(trace_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesGetTraceDetailResponse,
                    parse_obj_as(
                        type_=TracesGetTraceDetailResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def getanalyticssummary(
        self,
        *,
        start_time: str,
        end_time: str,
        function_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[TracesGetAnalyticsSummaryResponse]:
        """
        Parameters
        ----------
        start_time : str

        end_time : str

        function_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TracesGetAnalyticsSummaryResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            "traces/analytics",
            method="GET",
            params={
                "startTime": start_time,
                "endTime": end_time,
                "functionId": function_id,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesGetAnalyticsSummaryResponse,
                    parse_obj_as(
                        type_=TracesGetAnalyticsSummaryResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def listbyfunctionhash(
        self,
        hash: str,
        *,
        limit: typing.Optional[NumberFromString] = None,
        offset: typing.Optional[NumberFromString] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[TracesListByFunctionHashResponse]:
        """
        Parameters
        ----------
        hash : str

        limit : typing.Optional[NumberFromString]

        offset : typing.Optional[NumberFromString]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TracesListByFunctionHashResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            f"traces/function/hash/{jsonable_encoder(hash)}",
            method="GET",
            params={
                "limit": limit,
                "offset": offset,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesListByFunctionHashResponse,
                    parse_obj_as(
                        type_=TracesListByFunctionHashResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def searchbyenv(
        self,
        organization_id: str,
        project_id: str,
        environment_id: str,
        *,
        start_time: str,
        end_time: str,
        query: typing.Optional[str] = OMIT,
        input_messages_query: typing.Optional[str] = OMIT,
        output_messages_query: typing.Optional[str] = OMIT,
        fuzzy_search: typing.Optional[bool] = OMIT,
        trace_id: typing.Optional[str] = OMIT,
        span_id: typing.Optional[str] = OMIT,
        model: typing.Optional[typing.Sequence[str]] = OMIT,
        provider: typing.Optional[typing.Sequence[str]] = OMIT,
        function_id: typing.Optional[str] = OMIT,
        function_name: typing.Optional[str] = OMIT,
        span_name_prefix: typing.Optional[str] = OMIT,
        has_error: typing.Optional[bool] = OMIT,
        min_tokens: typing.Optional[float] = OMIT,
        max_tokens: typing.Optional[float] = OMIT,
        min_duration: typing.Optional[float] = OMIT,
        max_duration: typing.Optional[float] = OMIT,
        attribute_filters: typing.Optional[
            typing.Sequence[TracesSearchByEnvRequestAttributeFiltersItem]
        ] = OMIT,
        limit: typing.Optional[float] = OMIT,
        offset: typing.Optional[float] = OMIT,
        sort_by: typing.Optional[TracesSearchByEnvRequestSortBy] = OMIT,
        sort_order: typing.Optional[TracesSearchByEnvRequestSortOrder] = OMIT,
        root_spans_only: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[TracesSearchByEnvResponse]:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        environment_id : str

        start_time : str

        end_time : str

        query : typing.Optional[str]

        input_messages_query : typing.Optional[str]

        output_messages_query : typing.Optional[str]

        fuzzy_search : typing.Optional[bool]

        trace_id : typing.Optional[str]

        span_id : typing.Optional[str]

        model : typing.Optional[typing.Sequence[str]]

        provider : typing.Optional[typing.Sequence[str]]

        function_id : typing.Optional[str]

        function_name : typing.Optional[str]

        span_name_prefix : typing.Optional[str]

        has_error : typing.Optional[bool]

        min_tokens : typing.Optional[float]

        max_tokens : typing.Optional[float]

        min_duration : typing.Optional[float]

        max_duration : typing.Optional[float]

        attribute_filters : typing.Optional[typing.Sequence[TracesSearchByEnvRequestAttributeFiltersItem]]

        limit : typing.Optional[float]

        offset : typing.Optional[float]

        sort_by : typing.Optional[TracesSearchByEnvRequestSortBy]

        sort_order : typing.Optional[TracesSearchByEnvRequestSortOrder]

        root_spans_only : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TracesSearchByEnvResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/projects/{jsonable_encoder(project_id)}/environments/{jsonable_encoder(environment_id)}/traces/search",
            method="POST",
            json={
                "startTime": start_time,
                "endTime": end_time,
                "query": query,
                "inputMessagesQuery": input_messages_query,
                "outputMessagesQuery": output_messages_query,
                "fuzzySearch": fuzzy_search,
                "traceId": trace_id,
                "spanId": span_id,
                "model": model,
                "provider": provider,
                "functionId": function_id,
                "functionName": function_name,
                "spanNamePrefix": span_name_prefix,
                "hasError": has_error,
                "minTokens": min_tokens,
                "maxTokens": max_tokens,
                "minDuration": min_duration,
                "maxDuration": max_duration,
                "attributeFilters": convert_and_respect_annotation_metadata(
                    object_=attribute_filters,
                    annotation=typing.Sequence[
                        TracesSearchByEnvRequestAttributeFiltersItem
                    ],
                    direction="write",
                ),
                "limit": limit,
                "offset": offset,
                "sortBy": sort_by,
                "sortOrder": sort_order,
                "rootSpansOnly": root_spans_only,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesSearchByEnvResponse,
                    parse_obj_as(
                        type_=TracesSearchByEnvResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    def gettracedetailbyenv(
        self,
        organization_id: str,
        project_id: str,
        environment_id: str,
        trace_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[TracesGetTraceDetailByEnvResponse]:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        environment_id : str

        trace_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TracesGetTraceDetailByEnvResponse]
            Success
        """
        _response = self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/projects/{jsonable_encoder(project_id)}/environments/{jsonable_encoder(environment_id)}/traces/{jsonable_encoder(trace_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesGetTraceDetailByEnvResponse,
                    parse_obj_as(
                        type_=TracesGetTraceDetailByEnvResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )


class AsyncRawTracesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def create(
        self,
        *,
        resource_spans: typing.Sequence[TracesCreateRequestResourceSpansItem],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[TracesCreateResponse]:
        """
        Parameters
        ----------
        resource_spans : typing.Sequence[TracesCreateRequestResourceSpansItem]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TracesCreateResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            "traces",
            method="POST",
            json={
                "resourceSpans": convert_and_respect_annotation_metadata(
                    object_=resource_spans,
                    annotation=typing.Sequence[TracesCreateRequestResourceSpansItem],
                    direction="write",
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesCreateResponse,
                    parse_obj_as(
                        type_=TracesCreateResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 402:
                raise PaymentRequiredError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PlanLimitExceededError,
                        parse_obj_as(
                            type_=PlanLimitExceededError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def createotel(
        self,
        *,
        resource_spans: typing.Sequence[TracesCreateOtelRequestResourceSpansItem],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[TracesCreateOtelResponse]:
        """
        Parameters
        ----------
        resource_spans : typing.Sequence[TracesCreateOtelRequestResourceSpansItem]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TracesCreateOtelResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/traces",
            method="POST",
            json={
                "resourceSpans": convert_and_respect_annotation_metadata(
                    object_=resource_spans,
                    annotation=typing.Sequence[
                        TracesCreateOtelRequestResourceSpansItem
                    ],
                    direction="write",
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesCreateOtelResponse,
                    parse_obj_as(
                        type_=TracesCreateOtelResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 402:
                raise PaymentRequiredError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PlanLimitExceededError,
                        parse_obj_as(
                            type_=PlanLimitExceededError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def search(
        self,
        *,
        start_time: str,
        end_time: str,
        query: typing.Optional[str] = OMIT,
        input_messages_query: typing.Optional[str] = OMIT,
        output_messages_query: typing.Optional[str] = OMIT,
        fuzzy_search: typing.Optional[bool] = OMIT,
        trace_id: typing.Optional[str] = OMIT,
        span_id: typing.Optional[str] = OMIT,
        model: typing.Optional[typing.Sequence[str]] = OMIT,
        provider: typing.Optional[typing.Sequence[str]] = OMIT,
        function_id: typing.Optional[str] = OMIT,
        function_name: typing.Optional[str] = OMIT,
        span_name_prefix: typing.Optional[str] = OMIT,
        has_error: typing.Optional[bool] = OMIT,
        min_tokens: typing.Optional[float] = OMIT,
        max_tokens: typing.Optional[float] = OMIT,
        min_duration: typing.Optional[float] = OMIT,
        max_duration: typing.Optional[float] = OMIT,
        attribute_filters: typing.Optional[
            typing.Sequence[TracesSearchRequestAttributeFiltersItem]
        ] = OMIT,
        limit: typing.Optional[float] = OMIT,
        offset: typing.Optional[float] = OMIT,
        sort_by: typing.Optional[TracesSearchRequestSortBy] = OMIT,
        sort_order: typing.Optional[TracesSearchRequestSortOrder] = OMIT,
        root_spans_only: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[TracesSearchResponse]:
        """
        Parameters
        ----------
        start_time : str

        end_time : str

        query : typing.Optional[str]

        input_messages_query : typing.Optional[str]

        output_messages_query : typing.Optional[str]

        fuzzy_search : typing.Optional[bool]

        trace_id : typing.Optional[str]

        span_id : typing.Optional[str]

        model : typing.Optional[typing.Sequence[str]]

        provider : typing.Optional[typing.Sequence[str]]

        function_id : typing.Optional[str]

        function_name : typing.Optional[str]

        span_name_prefix : typing.Optional[str]

        has_error : typing.Optional[bool]

        min_tokens : typing.Optional[float]

        max_tokens : typing.Optional[float]

        min_duration : typing.Optional[float]

        max_duration : typing.Optional[float]

        attribute_filters : typing.Optional[typing.Sequence[TracesSearchRequestAttributeFiltersItem]]

        limit : typing.Optional[float]

        offset : typing.Optional[float]

        sort_by : typing.Optional[TracesSearchRequestSortBy]

        sort_order : typing.Optional[TracesSearchRequestSortOrder]

        root_spans_only : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TracesSearchResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            "traces/search",
            method="POST",
            json={
                "startTime": start_time,
                "endTime": end_time,
                "query": query,
                "inputMessagesQuery": input_messages_query,
                "outputMessagesQuery": output_messages_query,
                "fuzzySearch": fuzzy_search,
                "traceId": trace_id,
                "spanId": span_id,
                "model": model,
                "provider": provider,
                "functionId": function_id,
                "functionName": function_name,
                "spanNamePrefix": span_name_prefix,
                "hasError": has_error,
                "minTokens": min_tokens,
                "maxTokens": max_tokens,
                "minDuration": min_duration,
                "maxDuration": max_duration,
                "attributeFilters": convert_and_respect_annotation_metadata(
                    object_=attribute_filters,
                    annotation=typing.Sequence[TracesSearchRequestAttributeFiltersItem],
                    direction="write",
                ),
                "limit": limit,
                "offset": offset,
                "sortBy": sort_by,
                "sortOrder": sort_order,
                "rootSpansOnly": root_spans_only,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesSearchResponse,
                    parse_obj_as(
                        type_=TracesSearchResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def gettracedetail(
        self, trace_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[TracesGetTraceDetailResponse]:
        """
        Parameters
        ----------
        trace_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TracesGetTraceDetailResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"traces/{jsonable_encoder(trace_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesGetTraceDetailResponse,
                    parse_obj_as(
                        type_=TracesGetTraceDetailResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def getanalyticssummary(
        self,
        *,
        start_time: str,
        end_time: str,
        function_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[TracesGetAnalyticsSummaryResponse]:
        """
        Parameters
        ----------
        start_time : str

        end_time : str

        function_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TracesGetAnalyticsSummaryResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            "traces/analytics",
            method="GET",
            params={
                "startTime": start_time,
                "endTime": end_time,
                "functionId": function_id,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesGetAnalyticsSummaryResponse,
                    parse_obj_as(
                        type_=TracesGetAnalyticsSummaryResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def listbyfunctionhash(
        self,
        hash: str,
        *,
        limit: typing.Optional[NumberFromString] = None,
        offset: typing.Optional[NumberFromString] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[TracesListByFunctionHashResponse]:
        """
        Parameters
        ----------
        hash : str

        limit : typing.Optional[NumberFromString]

        offset : typing.Optional[NumberFromString]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TracesListByFunctionHashResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"traces/function/hash/{jsonable_encoder(hash)}",
            method="GET",
            params={
                "limit": limit,
                "offset": offset,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesListByFunctionHashResponse,
                    parse_obj_as(
                        type_=TracesListByFunctionHashResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def searchbyenv(
        self,
        organization_id: str,
        project_id: str,
        environment_id: str,
        *,
        start_time: str,
        end_time: str,
        query: typing.Optional[str] = OMIT,
        input_messages_query: typing.Optional[str] = OMIT,
        output_messages_query: typing.Optional[str] = OMIT,
        fuzzy_search: typing.Optional[bool] = OMIT,
        trace_id: typing.Optional[str] = OMIT,
        span_id: typing.Optional[str] = OMIT,
        model: typing.Optional[typing.Sequence[str]] = OMIT,
        provider: typing.Optional[typing.Sequence[str]] = OMIT,
        function_id: typing.Optional[str] = OMIT,
        function_name: typing.Optional[str] = OMIT,
        span_name_prefix: typing.Optional[str] = OMIT,
        has_error: typing.Optional[bool] = OMIT,
        min_tokens: typing.Optional[float] = OMIT,
        max_tokens: typing.Optional[float] = OMIT,
        min_duration: typing.Optional[float] = OMIT,
        max_duration: typing.Optional[float] = OMIT,
        attribute_filters: typing.Optional[
            typing.Sequence[TracesSearchByEnvRequestAttributeFiltersItem]
        ] = OMIT,
        limit: typing.Optional[float] = OMIT,
        offset: typing.Optional[float] = OMIT,
        sort_by: typing.Optional[TracesSearchByEnvRequestSortBy] = OMIT,
        sort_order: typing.Optional[TracesSearchByEnvRequestSortOrder] = OMIT,
        root_spans_only: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[TracesSearchByEnvResponse]:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        environment_id : str

        start_time : str

        end_time : str

        query : typing.Optional[str]

        input_messages_query : typing.Optional[str]

        output_messages_query : typing.Optional[str]

        fuzzy_search : typing.Optional[bool]

        trace_id : typing.Optional[str]

        span_id : typing.Optional[str]

        model : typing.Optional[typing.Sequence[str]]

        provider : typing.Optional[typing.Sequence[str]]

        function_id : typing.Optional[str]

        function_name : typing.Optional[str]

        span_name_prefix : typing.Optional[str]

        has_error : typing.Optional[bool]

        min_tokens : typing.Optional[float]

        max_tokens : typing.Optional[float]

        min_duration : typing.Optional[float]

        max_duration : typing.Optional[float]

        attribute_filters : typing.Optional[typing.Sequence[TracesSearchByEnvRequestAttributeFiltersItem]]

        limit : typing.Optional[float]

        offset : typing.Optional[float]

        sort_by : typing.Optional[TracesSearchByEnvRequestSortBy]

        sort_order : typing.Optional[TracesSearchByEnvRequestSortOrder]

        root_spans_only : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TracesSearchByEnvResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/projects/{jsonable_encoder(project_id)}/environments/{jsonable_encoder(environment_id)}/traces/search",
            method="POST",
            json={
                "startTime": start_time,
                "endTime": end_time,
                "query": query,
                "inputMessagesQuery": input_messages_query,
                "outputMessagesQuery": output_messages_query,
                "fuzzySearch": fuzzy_search,
                "traceId": trace_id,
                "spanId": span_id,
                "model": model,
                "provider": provider,
                "functionId": function_id,
                "functionName": function_name,
                "spanNamePrefix": span_name_prefix,
                "hasError": has_error,
                "minTokens": min_tokens,
                "maxTokens": max_tokens,
                "minDuration": min_duration,
                "maxDuration": max_duration,
                "attributeFilters": convert_and_respect_annotation_metadata(
                    object_=attribute_filters,
                    annotation=typing.Sequence[
                        TracesSearchByEnvRequestAttributeFiltersItem
                    ],
                    direction="write",
                ),
                "limit": limit,
                "offset": offset,
                "sortBy": sort_by,
                "sortOrder": sort_order,
                "rootSpansOnly": root_spans_only,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesSearchByEnvResponse,
                    parse_obj_as(
                        type_=TracesSearchByEnvResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )

    async def gettracedetailbyenv(
        self,
        organization_id: str,
        project_id: str,
        environment_id: str,
        trace_id: str,
        *,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[TracesGetTraceDetailByEnvResponse]:
        """
        Parameters
        ----------
        organization_id : str

        project_id : str

        environment_id : str

        trace_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TracesGetTraceDetailByEnvResponse]
            Success
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"organizations/{jsonable_encoder(organization_id)}/projects/{jsonable_encoder(project_id)}/environments/{jsonable_encoder(environment_id)}/traces/{jsonable_encoder(trace_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TracesGetTraceDetailByEnvResponse,
                    parse_obj_as(
                        type_=TracesGetTraceDetailByEnvResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        UnauthorizedErrorBody,
                        parse_obj_as(
                            type_=UnauthorizedErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        PermissionDeniedError,
                        parse_obj_as(
                            type_=PermissionDeniedError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        RateLimitError,
                        parse_obj_as(
                            type_=RateLimitError,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 500:
                raise InternalServerError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 503:
                raise ServiceUnavailableError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(
                status_code=_response.status_code,
                headers=dict(_response.headers),
                body=_response.text,
            )
        raise ApiError(
            status_code=_response.status_code,
            headers=dict(_response.headers),
            body=_response_json,
        )



import math

from mirascope import llm


@llm.tool
def sqrt_tool(number: float) -> float:
    """Computes the square root of a number"""
    return math.sqrt(number)


@llm.tool
def sum_tool(numbers: list[float]) -> float:
    total = 0
    for number in numbers:
        total += number
    return total


@llm.call("openai/gpt-5-mini", tools=[sqrt_tool, sum_tool])
def math_assistant(query: str):
    return query


response = math_assistant("What's the sum of the square roots of 137, 4242, and 6900?")

while response.tool_calls:
    tool_outputs = response.execute_tools()
    response = response.resume(tool_outputs)

print(response.pretty())
# sqrt(137) + sqrt(4242) + sqrt(6900) ≈ 159.9015764916355



"""Utilities for encoding Mirascope messages into GenAI semconv payloads."""

from __future__ import annotations

import json
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING

from . import gen_ai_types

if TYPE_CHECKING:
    from typing import TypeAlias

    MessagePart: TypeAlias = (
        gen_ai_types.TextPart
        | gen_ai_types.ToolCallRequestPart
        | gen_ai_types.ToolCallResponsePart
        | gen_ai_types.BlobPart
        | gen_ai_types.FilePart
        | gen_ai_types.UriPart
        | gen_ai_types.ReasoningPart
        | gen_ai_types.GenericPart
    )


from .....llm.content import (
    Audio,
    Base64ImageSource,
    Document,
    Image,
    Text,
    Thought,
    ToolCall,
    ToolOutput,
)
from .....llm.formatting import FormattableT
from .....llm.messages import AssistantMessage, Message, SystemMessage
from .....llm.responses.finish_reason import FinishReason
from .....llm.responses.root_response import RootResponse
from .....llm.tools import ToolkitT
from .....llm.types import Jsonable


@dataclass(frozen=True, slots=True)
class OTelMessageSnapshot:
    """Structured view of system, input, and output messages."""

    system_instructions: gen_ai_types.SystemInstructions
    """Instructions to be executed prior to the conversation."""

    inputs: gen_ai_types.InputMessages
    """Messages to be sent to the model."""

    outputs: gen_ai_types.OutputMessages
    """Messages received from the model."""


def _serialize_message_parts(
    parts: Sequence[
        Text | ToolCall | ToolOutput[Jsonable] | Thought | Image | Audio | Document
    ],
) -> list[MessagePart]:
    """Serialize message content parts into GenAI-compliant dictionaries.

    Handles Text, ToolCall, ToolOutput, Thought, Image, Audio, Document,
    and str types.
    """
    serialized: list[MessagePart] = []
    for part in parts:
        match part:
            case Text():
                text_part: gen_ai_types.TextPart = {
                    "type": "text",
                    "content": part.text,
                }
                serialized.append(text_part)
            case ToolCall():
                try:
                    arguments = json.loads(part.args)
                except json.JSONDecodeError:  # pragma: no cover
                    arguments = ""
                tool_call_part: gen_ai_types.ToolCallRequestPart = {
                    "type": "tool_call",
                    "id": part.id,
                    "name": part.name,
                    "arguments": arguments,
                }
                serialized.append(tool_call_part)
            case ToolOutput():
                tool_output_part: gen_ai_types.ToolCallResponsePart = {
                    "type": "tool_call_response",
                    "id": part.id,
                    "response": part.result,
                }
                serialized.append(tool_output_part)
            case Thought():
                serialized.append(
                    gen_ai_types.ReasoningPart(type="reasoning", content=part.thought)
                )
            case Image():
                if isinstance(part.source, Base64ImageSource):
                    image = gen_ai_types.BlobPart(
                        type="blob",
                        modality="image",
                        mime_type=part.source.mime_type,
                        content=part.source.data,
                    )
                else:
                    image = gen_ai_types.UriPart(
                        type="uri", modality="image", uri=part.source.url
                    )
                serialized.append(image)
            case Audio():
                serialized.append(
                    gen_ai_types.BlobPart(
                        type="blob",
                        modality="audio",
                        mime_type=part.source.mime_type,
                        content=part.source.data,
                    )
                )
            case Document():  # pragma: no cover
                raise NotImplementedError(
                    "Document serialization is not yet supported by any provider. "
                    "This will be implemented when provider support is added."
                )
    return serialized


def _serialize_message(message: Message) -> gen_ai_types.ChatMessage:
    """Serialize a Mirascope message into a GenAI ChatMessage."""
    content_parts = _serialize_message_parts(
        message.content if not isinstance(message, SystemMessage) else [message.content]
    )
    serialized: gen_ai_types.ChatMessage = {
        "role": message.role,
        "parts": content_parts,
    }
    name = getattr(message, "name", None)
    if name:
        serialized["name"] = name
    return serialized


def map_finish_reason(
    reason: FinishReason | None,
) -> gen_ai_types.FinishReason:
    """Map a finish reason to a GenAI finish reason."""
    if reason is None:
        return "stop"
    elif reason == FinishReason.MAX_TOKENS:
        return "length"
    elif reason == FinishReason.REFUSAL:  # pragma: no cover
        return "content_filter"
    raise ValueError(f"Unknown finish reason: {reason}")  # pragma: no cover


def _serialize_output_message(
    message: AssistantMessage, finish_reason: FinishReason | None
) -> gen_ai_types.OutputMessage:
    """Serialize an assistant message into a GenAI OutputMessage with finish_reason."""
    chat_message = _serialize_message(message)
    output_message: gen_ai_types.OutputMessage = {
        "role": chat_message["role"],
        "parts": chat_message["parts"],
        "finish_reason": map_finish_reason(finish_reason),
    }
    if "name" in chat_message:
        output_message["name"] = chat_message["name"]
    return output_message


def split_request_messages(
    messages: Sequence[Message],
) -> tuple[gen_ai_types.SystemInstructions, gen_ai_types.InputMessages]:
    """Return serialized system instructions and non-system input messages.

    System messages are flattened into a list of parts (not wrapped in message objects),
    while other messages are serialized as complete ChatMessage objects.
    """
    system_instructions: gen_ai_types.SystemInstructions = []
    inputs: gen_ai_types.InputMessages = []
    for message in messages:
        if message.role == "system":
            # System messages contribute only their parts (flattened)
            parts = _serialize_message_parts(
                [message.content]
                if isinstance(message, SystemMessage)
                else message.content
            )
            system_instructions.extend(parts)
        else:
            # Non-system messages are serialized as full ChatMessage objects
            serialized = _serialize_message(message)
            inputs.append(serialized)
    return system_instructions, inputs


def snapshot_from_response_messages(
    *,
    request_messages: Sequence[Message],
    assistant_message: AssistantMessage,
    finish_reason: FinishReason | None,
) -> OTelMessageSnapshot:
    """Build a snapshot using the request/response boundary for a call."""

    system_instructions, inputs = split_request_messages(request_messages)
    outputs = [_serialize_output_message(assistant_message, finish_reason)]
    return OTelMessageSnapshot(
        system_instructions=system_instructions,
        inputs=inputs,
        outputs=outputs,
    )


def snapshot_from_root_response(
    response: RootResponse[ToolkitT, FormattableT | None],
    *,
    request_messages: Sequence[Message] | None = None,
) -> OTelMessageSnapshot:
    """Build a snapshot directly from a `RootResponse`.

    Args:
        response: The response that includes the entire conversation history.
        request_messages: Optional explicit request message sequence. Defaults to all
            but the final assistant message inside `response.messages`.
    """

    assistant_message = response.messages[-1]
    if not isinstance(assistant_message, AssistantMessage):  # pragma: no cover
        raise TypeError("Final response message must be an AssistantMessage")

    return snapshot_from_response_messages(
        request_messages=request_messages or response.messages[:-1],
        assistant_message=assistant_message,
        finish_reason=response.finish_reason,
    )



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ...core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from ...core.serialization import FieldMetadata
from ...types.date import Date
from .organizations_subscription_response_scheduled_change_target_plan import (
    OrganizationsSubscriptionResponseScheduledChangeTargetPlan,
)


class OrganizationsSubscriptionResponseScheduledChange(UniversalBaseModel):
    target_plan: typing_extensions.Annotated[
        OrganizationsSubscriptionResponseScheduledChangeTargetPlan,
        FieldMetadata(alias="targetPlan"),
    ]
    effective_date: typing_extensions.Annotated[
        Date, FieldMetadata(alias="effectiveDate")
    ]
    schedule_id: typing_extensions.Annotated[str, FieldMetadata(alias="scheduleId")]

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
            extra="allow", frozen=True
        )  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



"""OpenTelemetry integration tests for `llm.Model.context_stream`."""

from __future__ import annotations

from collections.abc import Generator, Sequence
from typing import Any

import pytest
from inline_snapshot import snapshot
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export.in_memory_span_exporter import InMemorySpanExporter

from mirascope import llm, ops
from mirascope.ops._internal.configuration import set_tracer
from mirascope.ops._internal.instrumentation.llm import model as instrument_module
from tests.ops.utils import span_snapshot


@pytest.fixture(autouse=True, scope="function")
def initialize(tracer_provider: TracerProvider) -> Generator[None, None, None]:
    """Initialize ops configuration and LLM instrumentation for each test."""
    ops.configure(tracer_provider=tracer_provider)
    ops.instrument_llm()
    yield
    ops.uninstrument_llm()


@pytest.mark.vcr()
def test_model_context_stream_exports_genai_span(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Test that streaming a model call exports the correct OpenTelemetry span."""
    model = llm.Model(model_id="openai/gpt-4o-mini")
    ctx = llm.Context(deps={"tenant": "kai"})
    messages = [
        llm.messages.system("You are a concise assistant."),
        llm.messages.user("Say hello to the user named Kai."),
    ]

    response = model.context_stream(messages, ctx=ctx)
    response.finish()

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1
    span_dict = span_snapshot(spans[0])
    assert span_dict == snapshot(
        {
            "name": "chat openai/gpt-4o-mini",
            "kind": "CLIENT",
            "status": "UNSET",
            "attributes": {
                "gen_ai.operation.name": "chat",
                "gen_ai.provider.name": "openai",
                "gen_ai.request.model": "openai/gpt-4o-mini",
                "gen_ai.output.type": "text",
                "gen_ai.response.model": "openai/gpt-4o-mini",
                "gen_ai.response.finish_reasons": ["stop"],
                "gen_ai.system_instructions": '[{"type":"text","content":"You are a concise assistant."}]',
                "gen_ai.input.messages": '[{"role":"user","parts":[{"type":"text","content":"Say hello to the user named Kai."}]}]',
                "gen_ai.output.messages": '[{"role":"assistant","parts":[{"type":"text","content":"Hello, Kai!"}],"finish_reason":"stop"}]',
                "gen_ai.usage.input_tokens": 25,
                "gen_ai.usage.output_tokens": 5,
                "mirascope.response.messages": '[{"role":"system","content":{"type":"text","text":"You are a concise assistant."}},{"role":"user","content":[{"type":"text","text":"Say hello to the user named Kai."}],"name":null}]',
                "mirascope.response.content": '[{"type":"text","text":"Hello, Kai!"}]',
                "mirascope.response.usage": '{"input_tokens":25,"output_tokens":5,"cache_read_tokens":0,"cache_write_tokens":0,"reasoning_tokens":0,"total_tokens":30}',
            },
        }
    )


def test_model_context_stream_without_tracer_returns_response(
    span_exporter: InMemorySpanExporter, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test that streaming a model call returns the response object."""
    set_tracer(None)
    dummy_response = object()

    def _fake_context_stream(
        self: llm.Model,
        content: llm.UserContent | Sequence[llm.Message],
        *,
        ctx: llm.Context[Any],
        tools: object | None = None,
        format: object | None = None,
    ) -> object:
        return dummy_response

    monkeypatch.setattr(
        instrument_module, "_ORIGINAL_MODEL_CONTEXT_STREAM", _fake_context_stream
    )

    model = llm.Model(model_id="openai/gpt-4o-mini")
    ctx = llm.Context(deps={})
    messages = [llm.messages.user("hi")]

    response = model.context_stream(messages, ctx=ctx)

    assert response is dummy_response
    assert span_exporter.get_finished_spans() == ()


def test_model_context_stream_records_error_on_exception(
    span_exporter: InMemorySpanExporter, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Test that streaming a model call exports the correct OpenTelemetry span."""

    def _failing_context_stream(
        self: llm.Model,
        content: Sequence[llm.Message],
        *,
        ctx: llm.Context[Any],
        tools: object | None = None,
        format: object | None = None,
    ) -> object:
        raise ValueError("error")

    monkeypatch.setattr(
        instrument_module, "_ORIGINAL_MODEL_CONTEXT_STREAM", _failing_context_stream
    )

    model = llm.Model(model_id="openai/gpt-4o-mini")
    ctx = llm.Context(deps={})
    messages = [llm.messages.user("hi")]

    with pytest.raises(ValueError, match="error"):
        model.context_stream(messages, ctx=ctx)

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1
    span_dict = span_snapshot(spans[0])
    assert span_dict == snapshot(
        {
            "name": "chat openai/gpt-4o-mini",
            "kind": "CLIENT",
            "status": "ERROR",
            "attributes": {
                "gen_ai.operation.name": "chat",
                "gen_ai.provider.name": "openai",
                "gen_ai.request.model": "openai/gpt-4o-mini",
                "gen_ai.output.type": "text",
                "gen_ai.input.messages": '[{"role":"user","parts":[{"type":"text","content":"hi"}]}]',
                "error.type": "ValueError",
                "error.message": "error",
            },
        }
    )



from mirascope import llm


def recommend_book(genre: str) -> llm.StreamResponse:
    model: llm.Model = llm.use_model("openai/gpt-5")
    return model.stream(f"Please recommend a book in {genre}.")


def main():
    response: llm.StreamResponse = recommend_book("fantasy")
    for chunk in response.pretty_stream():
        print(chunk, flush=True, end="")


main()



"""Base class for StreamResponse and AsyncStreamResponse."""

from collections.abc import AsyncIterator, Iterator, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeAlias, TypeVar

from ..content import (
    AssistantContentChunk,
    AssistantContentPart,
    Text,
    TextChunk,
    TextEndChunk,
    TextStartChunk,
    Thought,
    ThoughtChunk,
    ThoughtEndChunk,
    ThoughtStartChunk,
    ToolCall,
    ToolCallChunk,
    ToolCallEndChunk,
    ToolCallStartChunk,
)
from ..formatting import (
    Format,
    FormattableT,
    Partial,
    is_output_parser,
)
from ..messages import AssistantMessage, Message
from ..tools import FORMAT_TOOL_NAME, ToolkitT
from ..types import Jsonable
from .finish_reason import FinishReasonChunk
from .root_response import RootResponse
from .streams import (
    AsyncStream,
    AsyncTextStream,
    AsyncThoughtStream,
    AsyncToolCallStream,
    Stream,
    TextStream,
    ThoughtStream,
    ToolCallStream,
)
from .usage import Usage, UsageDeltaChunk

if TYPE_CHECKING:
    from ..models import Params
    from ..providers import ModelId, ProviderId


StreamResponseT = TypeVar("StreamResponseT", bound="BaseStreamResponse[Any, Any, Any]")


@dataclass(kw_only=True)
class RawStreamEventChunk:
    """A chunk containing a raw stream event from the underlying provider.

    Will be accumulated on `StreamResponse.raw` for debugging purposes.
    """

    type: Literal["raw_stream_event_chunk"] = "raw_stream_event_chunk"

    raw_stream_event: Any
    """The raw stream event from the underlying provider."""


@dataclass(kw_only=True)
class RawMessageChunk:
    """A chunk containing provider-specific raw message content that will be added to the `AssistantMessage`.

    This chunk contains a provider-specific representation of a piece of content that
    will be added to the `AssistantMessage` reconstructed by the containing stream.
    This content should be a Jsonable Python object for serialization purposes.

    The intention is that this content may be passed as-is back to the provider when the
    generated `AssistantMessage` is being reused in conversation.
    """

    type: Literal["raw_message_chunk"] = "raw_message_chunk"

    raw_message: Jsonable
    """The provider-specific raw content.
    
    Should be a Jsonable object.
    """


StreamResponseChunk: TypeAlias = (
    AssistantContentChunk
    | FinishReasonChunk
    | RawStreamEventChunk
    | RawMessageChunk
    | UsageDeltaChunk
)

ChunkIterator: TypeAlias = Iterator[StreamResponseChunk]
"""Synchronous iterator yielding chunks with raw data."""

AsyncChunkIterator: TypeAlias = AsyncIterator[StreamResponseChunk]
"""Asynchronous iterator yielding chunks with raw data."""

ChunkIteratorT = TypeVar("ChunkIteratorT", bound=ChunkIterator | AsyncChunkIterator)


class BaseStreamResponse(
    RootResponse[ToolkitT, FormattableT],
    Generic[ChunkIteratorT, ToolkitT, FormattableT],
):
    """Base class underpinning StreamResponse and AsyncStreamResponse.

    Manages chunk handling logic for both.
    """

    raw_stream_events: Sequence[Any]
    """The raw stream event chunks from the LLM. Provider-specific."""

    chunks: Sequence[AssistantContentChunk]
    """All of the Mirascope chunks consumed from the stream."""

    content: Sequence[AssistantContentPart]
    """The content generated by the LLM.
    
    Content is updated in this array as it is consumed by the stream. Text content will 
    update with each text chunk (this will mutate the Text object that is returned 
    rather than creating a new one). Other content will be added once each part
    is fully streamed.
    """

    messages: list[Message]
    """The message history, including the most recent assistant message.

    The most recent assistant message will have all of the completed content that has 
    already been consumed from the stream. Text content will be included as each chunk 
    is processed; other content will be included only when its corresponding part is 
    completed (to avoid partial tool calls and the like). If no content has been 
    streamed, then the final assistant message will be present (to maintain turn order 
    expectations), but will be empty.
    """

    texts: Sequence[Text]
    """The text content in the generated response, if any.
    
    Text content updates with each text chunk as it streams. The `Text` objects are 
    mutated in place rather than creating new ones for each chunk.
    """

    tool_calls: Sequence[ToolCall]
    """The tools the LLM wants called on its behalf, if any.
    
    Tool calls are only added to this sequence once they have been fully streamed 
    to avoid partial tool calls in the response.
    """

    thoughts: Sequence[Thought]
    """The readable thoughts from the model's thinking process, if any.
    
    The thoughts may be direct output from the model thinking process, or may be a
    generated summary. (This depends on the provider; newer models tend to summarize.)

    Thoughts are added to the sequence as they are streamed. The `Thought` objects are
    mutated in place rather than creating new ones for each chunk.
    """

    consumed: bool = False
    """Whether the stream has been fully consumed.
    
    This is True after all chunks have been processed from the underlying iterator.
    When False, more content may be available by calling the stream methods.
    """

    def __init__(
        self,
        *,
        provider_id: "ProviderId",
        model_id: "ModelId",
        provider_model_name: str,
        params: "Params",
        toolkit: ToolkitT,
        format: Format[FormattableT] | None = None,
        input_messages: Sequence[Message],
        chunk_iterator: ChunkIteratorT,
        usage: Usage | None = None,
    ) -> None:
        """Initialize the BaseStreamResponse.

        Args:
            provider: The provider name (e.g. "anthropic", "openai").
            model_id: The model identifier that generated the response.
            provider_model_name: Optional provider-specific model name. May include
                provider-specific additional info (like api mode in "gpt-5:responses").
            params: The params used to generate the response (or None).
            toolkit: Toolkit containing all the tools used to generate the response.
            format: The `Format` for the expected structured output format (or None).
            input_messages: The input messages that were sent to the LLM
            usage: Token usage statistics for the response.

        The BaseStreamResponse will process the tuples to build the chunks and raw lists
        as the stream is consumed.
        """

        self.provider_id = provider_id
        self.model_id = model_id
        self.provider_model_name = provider_model_name
        self.params = params
        self.toolkit = toolkit
        self.usage = usage
        self.format = format

        # Internal-only lists which we mutate (append) during chunk processing
        self._chunks: list[AssistantContentChunk] = []
        self._content: list[AssistantContentPart] = []
        self._texts: list[Text] = []
        self._thoughts: list[Thought] = []
        self._tool_calls: list[ToolCall] = []
        self._raw_stream_events: list[Any] = []
        self._last_raw_stream_event_chunk: Any | None = None

        # Externally-facing references typed as immutable Sequences
        self.chunks = self._chunks
        self.content = self._content
        self.texts = self._texts
        self.thoughts = self._thoughts
        self.tool_calls = self._tool_calls
        self.raw_stream_events = self._raw_stream_events

        self.finish_reason = None

        self._assistant_message = AssistantMessage(
            content=self._content,
            provider_id=provider_id,
            model_id=model_id,
            provider_model_name=provider_model_name,
            raw_message=None,
        )

        self.messages = list(input_messages) + [self._assistant_message]

        self._chunk_iterator = chunk_iterator
        self._current_content: Text | Thought | None = None
        self._current_tool_calls: dict[str, ToolCall] = {}

        self._processing_format_tool: bool = False

    def _transform_format_tool_chunks(
        self, chunk: AssistantContentChunk
    ) -> AssistantContentChunk:
        if chunk.type == "tool_call_start_chunk" and chunk.name.startswith(
            FORMAT_TOOL_NAME
        ):
            self._processing_format_tool = True
            return TextStartChunk()
        if self._processing_format_tool and chunk.type == "tool_call_chunk":
            return TextChunk(delta=chunk.delta)
        if self._processing_format_tool and chunk.type == "tool_call_end_chunk":
            self._processing_format_tool = False
            return TextEndChunk()
        return chunk

    def _handle_chunk(self, chunk: AssistantContentChunk) -> AssistantContentChunk:
        if self.finish_reason:
            raise RuntimeError(
                f"Stream already finished with reason: {self.finish_reason}"
            )
        chunk = self._transform_format_tool_chunks(chunk)

        if chunk.content_type == "text":
            self._handle_text_chunk(chunk)
        elif chunk.content_type == "tool_call":
            self._handle_tool_call_chunk(chunk)
        elif chunk.content_type == "thought":
            self._handle_thought_chunk(chunk)
        else:
            raise NotImplementedError

        self._chunks.append(chunk)
        return chunk

    def _handle_text_chunk(
        self, chunk: TextStartChunk | TextChunk | TextEndChunk
    ) -> None:
        if chunk.type == "text_start_chunk":
            if self._current_content or self._current_tool_calls:
                raise RuntimeError(
                    "Received text_start_chunk while processing another chunk"
                )
            self._current_content = Text(text="")
            # Text gets included in content even when unfinished.
            self._content.append(self._current_content)
            self._texts.append(self._current_content)

        elif chunk.type == "text_chunk":
            if self._current_content is None or self._current_content.type != "text":
                raise RuntimeError("Received text_chunk while not processing text.")
            self._current_content.text += chunk.delta

        elif chunk.type == "text_end_chunk":
            if self._current_content is None or self._current_content.type != "text":
                raise RuntimeError("Received text_end_chunk while not processing text.")
            self._current_content = None

    def _handle_thought_chunk(
        self, chunk: ThoughtStartChunk | ThoughtChunk | ThoughtEndChunk
    ) -> None:
        if chunk.type == "thought_start_chunk":
            if self._current_content or self._current_tool_calls:
                raise RuntimeError(
                    "Received thought_start_chunk while processing another chunk"
                )
            self._current_content = Thought(thought="")
            # Thoughts get included even when unfinished.
            self._content.append(self._current_content)
            self._thoughts.append(self._current_content)

        elif chunk.type == "thought_chunk":
            if self._current_content is None or self._current_content.type != "thought":
                raise RuntimeError(
                    "Received thought_chunk while not processing thought."
                )
            self._current_content.thought += chunk.delta

        elif chunk.type == "thought_end_chunk":
            if self._current_content is None or self._current_content.type != "thought":
                raise RuntimeError(
                    "Received thought_end_chunk while not processing thought."
                )
            self._current_content = None

    def _handle_tool_call_chunk(
        self, chunk: ToolCallStartChunk | ToolCallChunk | ToolCallEndChunk
    ) -> None:
        if chunk.type == "tool_call_start_chunk":
            if self._current_content:
                raise RuntimeError(
                    "Received tool_call_start_chunk while processing another chunk"
                )
            if chunk.id in self._current_tool_calls:
                raise RuntimeError("Got tool_call_start_chunk with conflicting id")
            # Create a new tool call and track it by ID
            # Multiple tool calls can be in progress simultaneously (interleaved)
            tool_call = ToolCall(
                id=chunk.id,
                name=chunk.name,
                args="",
            )
            self._current_tool_calls[chunk.id] = tool_call

        elif chunk.type == "tool_call_chunk":
            # Look up the tool call by ID
            tool_call = self._current_tool_calls.get(chunk.id)
            if tool_call is None:
                raise RuntimeError(
                    f"Received tool_call_chunk for unknown tool call ID: {chunk.id}"
                )
            tool_call.args += chunk.delta

        elif chunk.type == "tool_call_end_chunk":
            # Finalize the tool call
            tool_call = self._current_tool_calls.get(chunk.id)
            if tool_call is None:
                raise RuntimeError(
                    f"Received tool_call_end_chunk for unknown tool call ID: {chunk.id}"
                )
            if not tool_call.args:
                tool_call.args = "{}"
            self._content.append(tool_call)
            self._tool_calls.append(tool_call)
            del self._current_tool_calls[chunk.id]

    def _pretty_chunk(self, chunk: AssistantContentChunk, spacer: str) -> str:
        match chunk.type:
            case "text_start_chunk":
                return spacer
            case "text_chunk":
                return chunk.delta
            case "tool_call_start_chunk":
                return spacer + f"**ToolCall ({chunk.name}):** "
            case "tool_call_chunk":
                return chunk.delta
            case "thought_start_chunk":
                return spacer + "**Thinking:**\n  "
            case "thought_chunk":
                return chunk.delta.replace("\n", "\n  ")  # Indent every line
            case _:
                return ""


class BaseSyncStreamResponse(BaseStreamResponse[ChunkIterator, ToolkitT, FormattableT]):
    """A base class for synchronous Stream Responses."""

    def streams(self) -> Iterator[Stream]:
        """Returns an iterator that yields streams for each content part in the response.

        Returns:
            Iterator[Stream]: Synchronous iterator yielding Stream objects

        Each content part in the response will correspond to one stream, which will yield
        chunks of content as they come in from the underlying LLM.

        Fully iterating through this iterator will fully consume the underlying stream,
        updating the Response with all collected content.

        As content is consumed, it is cached on the StreamResponse. If a new iterator
        is constructed via calling `streams()`, it will start by replaying the cached
        content from the response, and (if there is still more content to consume from
        the LLM), it will proceed to consume it once it has iterated through all the
        cached chunks.
        """
        chunk_iter = self.chunk_stream()

        for start_chunk in chunk_iter:
            # At the start of this loop, we always expect to find a start chunk. Then,
            # before proceeding, we will collect from the stream we create (in case the
            # user did not exhaust it), which ensures we will be expecting a start chunk
            # again on the next iteration
            match start_chunk.type:
                case "text_start_chunk":

                    def text_stream_iterator() -> Iterator[TextChunk]:
                        for chunk in chunk_iter:
                            if chunk.type == "text_chunk":
                                yield chunk
                            else:
                                return  # Stream finished

                    stream = TextStream(chunk_iterator=text_stream_iterator())
                    yield stream

                case "thought_start_chunk":

                    def thought_stream_iterator() -> Iterator[ThoughtChunk]:
                        for chunk in chunk_iter:
                            if chunk.type == "thought_chunk":
                                yield chunk
                            else:
                                return  # Stream finished

                    stream = ThoughtStream(chunk_iterator=thought_stream_iterator())
                    yield stream

                case "tool_call_start_chunk":
                    tool_id = start_chunk.id
                    tool_name = start_chunk.name

                    def tool_call_stream_iterator() -> Iterator[ToolCallChunk]:
                        for chunk in chunk_iter:
                            if chunk.type == "tool_call_chunk":
                                yield chunk
                            else:
                                return  # Stream finished

                    stream = ToolCallStream(
                        tool_id=tool_id,
                        tool_name=tool_name,
                        chunk_iterator=tool_call_stream_iterator(),
                    )
                    yield stream

                case _:  # pragma: no cover
                    raise RuntimeError(f"Expected start chunk, got: {start_chunk.type}")

            # Before continuing to the next stream, make sure the last stream is consumed
            # (If the user did not do so when we yielded it)
            stream.collect()

    def chunk_stream(
        self,
    ) -> Iterator[AssistantContentChunk]:
        """Returns an iterator that yields content chunks as they are received.

        Returns:
            Iterator[AssistantContentChunk]: Synchronous iterator yielding chunks

        This provides access to the Mirascope chunk data including start, delta, and end chunks
        for each content type (text, thought, tool_call). Unlike the streams() method
        that groups chunks by content part, this yields individual chunks as they arrive.

        Fully iterating through this iterator will fully consume the underlying stream,
        updating the Response with all collected content.

        As chunks are consumed, they are cached on the StreamResponse. If a new iterator
        is constructed via calling `chunk_stream()`, it will start by replaying the cached
        chunks from the response, and (if there is still more content to consume from
        the LLM), it will proceed to consume it once it has iterated through all the
        cached chunks.
        """
        for chunk in self.chunks:
            yield chunk

        if self.consumed:
            return

        for chunk in self._chunk_iterator:
            if chunk.type == "raw_stream_event_chunk":
                self._raw_stream_events.append(chunk.raw_stream_event)
            elif chunk.type == "raw_message_chunk":
                self._assistant_message.raw_message = chunk.raw_message
            elif chunk.type == "finish_reason_chunk":
                self.finish_reason = chunk.finish_reason
            elif chunk.type == "usage_delta_chunk":
                if self.usage is None:
                    self.usage = Usage()
                self.usage.input_tokens += chunk.input_tokens
                self.usage.output_tokens += chunk.output_tokens
                self.usage.cache_read_tokens += chunk.cache_read_tokens
                self.usage.cache_write_tokens += chunk.cache_write_tokens
                self.usage.reasoning_tokens += chunk.reasoning_tokens
                if chunk.provider_tool_usage:
                    self.usage.provider_tool_usage = chunk.provider_tool_usage
            else:
                yield self._handle_chunk(chunk)

        self.consumed = True

    def finish(self) -> None:
        """Finish streaming all of this response's content."""
        for _ in self.chunk_stream():
            pass

    def text_stream(self, sep: str = "\n") -> Iterator[str]:
        """Stream only the text content from the response.

        Args:
            sep: Separator to yield between text parts. Defaults to newline.

        Returns:
            Iterator[str]: Iterator yielding text delta strings

        Yields text deltas as they arrive, ignoring thoughts, tool calls, and other
        content types. Ideal for displaying text to users in real-time.

        If you concatenate the strings from `.text_stream()`, it will be equivalent to
        calling `.text(sep=sep)` on the fully consumed response.
        """
        for stream in self.streams():
            if stream.content_type == "text":
                yield from stream
                yield sep

    def pretty_stream(self) -> Iterator[str]:
        """Stream a readable representation of the stream_response as text.

        Returns:
            Iterator[str]: Iterator yielding string chunks depicting the content

        Iterating through the pretty stream will populate the stream response by consuming
        the underlying iterator (if it hasn't been consumed already). Calling `.pretty_stream()`
        will always return a fresh iterator that begins from the start of the stream.

        If you concatenate the text from `.pretty_stream()`, it will be equivalent to the
        text generated by calling `.pretty()` (assuming the response was fully consumed
        at the time when you call `.pretty()`).
        """
        printed = False

        for chunk in self.chunk_stream():
            pretty = self._pretty_chunk(chunk, "\n\n" if printed else "")
            if pretty != "":
                printed = True
            yield pretty

    def structured_stream(
        self,
    ) -> Iterator[Partial[FormattableT]]:
        """Drive the stream forward, yielding partial formatted outputs as they arrive.

        This method consumes the underlying stream chunk by chunk, yielding parsed
        partial outputs each time new content arrives. Each yielded value is a
        Partial[FormattableT] with optional fields that may be None until fully received.

        Example:
            >>> response = recommend_book.stream("fantasy")
            >>> for partial in response.structured_stream():
            >>>     print(f"Title so far: {partial.title}")
            >>> book = response.parse()  # Get final complete result

        Fully iterating through this iterator will fully consume the underlying stream,
        updating the Response with all collected content.

        Yields:
            Partial[FormattableT]: Partial objects with fields populated as they arrive.
                Fields not yet received will be None.

        Raises:
            ValueError: If format parameter not set.
            NotImplementedError: If format uses OutputParser (not supported).
        """
        if self.format is None:
            raise ValueError("structured_stream() requires format parameter")

        if is_output_parser(self.format.formattable):
            raise NotImplementedError(
                "structured_stream() not supported for OutputParser. "
                "Use BaseModel or primitive types."
            )

        for chunk in self.chunk_stream():
            if chunk.type == "text_chunk":
                partial = self.parse(partial=True)
                if partial:
                    yield partial


class BaseAsyncStreamResponse(
    BaseStreamResponse[AsyncChunkIterator, ToolkitT, FormattableT]
):
    """A base class for asynchronous Stream Responses."""

    async def streams(self) -> AsyncIterator[AsyncStream]:
        """Returns an async iterator that yields streams for each content part in the response.

        Returns:
            AsyncIterator[AsyncStream]: Async iterator yielding AsyncStream objects

        Each content part in the response will correspond to one stream, which will yield
        chunks of content as they come in from the underlying LLM.

        Fully iterating through this iterator will fully consume the underlying stream,
        updating the Response with all collected content.

        As content is consumed, it is cached on the AsyncStreamResponse. If a new iterator
        is constructed via calling `streams()`, it will start by replaying the cached
        content from the response, and (if there is still more content to consume from
        the LLM), it will proceed to consume it once it has iterated through all the
        cached chunks.
        """
        chunk_iter = self.chunk_stream()

        async for start_chunk in chunk_iter:
            # At the start of this loop, we always expect to find a start chunk. Then,
            # before proceeding, we will collect from the stream we create (in case the
            # user did not exhaust it), which ensures we will be expecting a start chunk
            # again on the next iteration
            match start_chunk.type:
                case "text_start_chunk":

                    async def text_stream_iterator() -> AsyncIterator[TextChunk]:
                        async for chunk in chunk_iter:
                            if chunk.type == "text_chunk":
                                yield chunk
                            else:
                                return  # Stream finished

                    stream = AsyncTextStream(chunk_iterator=text_stream_iterator())
                    yield stream

                case "thought_start_chunk":

                    async def thought_stream_iterator() -> AsyncIterator[ThoughtChunk]:
                        async for chunk in chunk_iter:
                            if chunk.type == "thought_chunk":
                                yield chunk
                            else:
                                return  # Stream finished

                    stream = AsyncThoughtStream(
                        chunk_iterator=thought_stream_iterator()
                    )
                    yield stream

                case "tool_call_start_chunk":
                    tool_id = start_chunk.id
                    tool_name = start_chunk.name

                    async def tool_call_stream_iterator() -> AsyncIterator[
                        ToolCallChunk
                    ]:
                        async for chunk in chunk_iter:
                            if chunk.type == "tool_call_chunk":
                                yield chunk
                            else:
                                return  # Stream finished

                    stream = AsyncToolCallStream(
                        tool_id=tool_id,
                        tool_name=tool_name,
                        chunk_iterator=tool_call_stream_iterator(),
                    )
                    yield stream

                case _:  # pragma: no cover
                    raise RuntimeError(f"Expected start chunk, got: {start_chunk.type}")

            # Before continuing to the next stream, make sure the last stream is consumed
            # (If the user did not do so when we yielded it)
            await stream.collect()

    async def chunk_stream(
        self,
    ) -> AsyncIterator[AssistantContentChunk]:
        """Returns an async iterator that yields content chunks as they are received.

        Returns:
            AsyncIterator[AssistantContentChunk]: Async iterator yielding chunks

        This provides access to the Mirascope chunk data including start, delta, and end chunks
        for each content type (text, thinking, tool_call). Unlike the streams() method
        that groups chunks by content part, this yields individual chunks as they arrive.

        Fully iterating through this iterator will fully consume the underlying stream,
        updating the Response with all collected content.

        As chunks are consumed, they are cached on the AsyncStreamResponse. If a new iterator
        is constructed via calling `chunk_stream()`, it will start by replaying the cached
        chunks from the response, and (if there is still more content to consume from
        the LLM), it will proceed to consume it once it has iterated through all the
        cached chunks.
        """

        for chunk in self.chunks:
            yield chunk

        if self.consumed:
            return

        async for chunk in self._chunk_iterator:
            if chunk.type == "raw_stream_event_chunk":
                self._raw_stream_events.append(chunk.raw_stream_event)
            elif chunk.type == "raw_message_chunk":
                self._assistant_message.raw_message = chunk.raw_message
            elif chunk.type == "finish_reason_chunk":
                self.finish_reason = chunk.finish_reason
            elif chunk.type == "usage_delta_chunk":
                if self.usage is None:
                    self.usage = Usage()
                self.usage.input_tokens += chunk.input_tokens
                self.usage.output_tokens += chunk.output_tokens
                self.usage.cache_read_tokens += chunk.cache_read_tokens
                self.usage.cache_write_tokens += chunk.cache_write_tokens
                self.usage.reasoning_tokens += chunk.reasoning_tokens
                if chunk.provider_tool_usage:
                    self.usage.provider_tool_usage = chunk.provider_tool_usage
            else:
                yield self._handle_chunk(chunk)

        self.consumed = True

    async def finish(self) -> None:
        """Finish streaming all of this response's content."""
        async for _ in self.chunk_stream():
            pass

    async def text_stream(self, sep: str = "\n") -> AsyncIterator[str]:
        """Stream only the text content from the response.

        Args:
            sep: Separator to yield between text parts. Defaults to newline.

        Returns:
            AsyncIterator[str]: Async iterator yielding text delta strings

        Yields text deltas as they arrive, ignoring thoughts, tool calls, and other
        content types. Ideal for displaying text to users in real-time.

        If you concatenate the strings from `.text_stream()`, it will be equivalent to
        calling `.text(sep=sep)` on the fully consumed response.
        """
        async for stream in self.streams():
            if stream.content_type == "text":
                async for delta in stream:
                    yield delta
                yield sep

    async def pretty_stream(self) -> AsyncIterator[str]:
        """Stream a readable representation of the stream_response as text.

        Returns:
            AsyncIterator[str]: Async iterator yielding string chunks depicting the content

        Iterating through the pretty stream will populate the stream response by consuming
        the underlying iterator (if it hasn't been consumed already). Calling `.pretty_stream()`
        will always return a fresh iterator that begins from the start of the stream.

        If you concatenate the text from `.pretty_stream()`, it will be equivalent to the
        text generated by calling `.pretty()` (assuming the response was fully consumed
        at the time when you call `.pretty()`).
        """
        printed = False

        async for chunk in self.chunk_stream():
            pretty = self._pretty_chunk(chunk, "\n\n" if printed else "")
            if pretty != "":
                printed = True
            yield pretty

    async def structured_stream(
        self,
    ) -> AsyncIterator[Partial[FormattableT]]:
        """Drive the stream forward, yielding partial formatted outputs as they arrive.

        This method consumes the underlying stream chunk by chunk, yielding parsed
        partial outputs each time new content arrives. Each yielded value is a
        Partial[FormattableT] with optional fields that may be None until fully received.

        Example:
            >>> response = await recommend_book.stream("fantasy")
            >>> async for partial in response.structured_stream():
            >>>     print(f"Title so far: {partial.title}")
            >>> book = response.parse()  # Get final complete result

        Fully iterating through this iterator will fully consume the underlying stream,
        updating the Response with all collected content.

        Yields:
            Partial[FormattableT]: Partial objects with fields populated as they arrive.
                Fields not yet received will be None.

        Raises:
            ValueError: If format parameter not set.
            NotImplementedError: If format uses OutputParser (not supported).
        """
        if self.format is None:
            raise ValueError("structured_stream() requires format parameter")

        if is_output_parser(self.format.formattable):
            raise NotImplementedError(
                "structured_stream() not supported for OutputParser. "
                "Use BaseModel or primitive types."
            )

        async for chunk in self.chunk_stream():
            if chunk.type == "text_chunk":
                partial = self.parse(partial=True)
                if partial:
                    yield partial



from pydantic import BaseModel

from mirascope import llm


class Book(BaseModel):
    title: str
    author: str
    summary: str


@llm.call("openai/gpt-5-mini", format=Book)
def recommend_book(genre: str):
    return f"Recommend a {genre} book."


response = recommend_book.stream("fantasy")

# Stream partial objects as fields are populated
for partial in response.structured_stream():
    # partial is a Partial[Book] with all fields optional
    print(f"Partial: {partial}")

# Get the final validated object
book = response.parse()
print(f"\nFinal: {book.title} by {book.author}")



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    FinishReason,
    Text,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider_id": "mlx",
            "model_id": "mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
            "provider_model_name": "mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
            "params": {"max_tokens": 50},
            "finish_reason": FinishReason.MAX_TOKENS,
            "usage": {
                "input_tokens": 15,
                "output_tokens": 50,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "MLXCassetteResponse(text=' start', token=1191, prompt_tokens=15, generation_tokens=50, finish_reason='length')",
                "total_tokens": 65,
            },
            "messages": [
                UserMessage(content=[Text(text="List all U.S. states.")]),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
<think>
Okay, the user is asking me to list all U.S. states. Let me start by recalling the main states. I know that the United States has 50 states, so I should list them all. First, I'll start\
"""
                        )
                    ],
                    provider_id="mlx",
                    model_id="mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
                    provider_model_name=None,
                    raw_message=None,
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider_id": "mlx",
            "model_id": "mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
            "provider_model_name": "mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
            "params": {"max_tokens": 50},
            "finish_reason": FinishReason.MAX_TOKENS,
            "usage": {
                "input_tokens": 15,
                "output_tokens": 50,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "MLXCassetteResponse(text=' start', token=1191, prompt_tokens=15, generation_tokens=50, finish_reason='length')",
                "total_tokens": 65,
            },
            "messages": [
                UserMessage(content=[Text(text="List all U.S. states.")]),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
<think>
Okay, the user is asking me to list all U.S. states. Let me start by recalling the main states. I know that the United States has 50 states, so I should list them all. First, I'll start\
"""
                        )
                    ],
                    provider_id="mlx",
                    model_id="mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
                    provider_model_name=None,
                    raw_message=None,
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "mlx",
            "model_id": "mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
            "provider_model_name": "mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
            "finish_reason": FinishReason.MAX_TOKENS,
            "messages": [
                UserMessage(content=[Text(text="List all U.S. states.")]),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
<think>
Okay, the user is asking me to list all U.S. states. Let me start by recalling the main states. I know that the United States has 50 states, so I should list them all. First, I'll start\
"""
                        )
                    ],
                    provider_id="mlx",
                    model_id="mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
                    provider_model_name="mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
                    raw_message=None,
                ),
            ],
            "format": None,
            "tools": [],
            "usage": {
                "input_tokens": 15,
                "output_tokens": 50,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 65,
            },
            "n_chunks": 51,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "mlx",
            "model_id": "mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
            "provider_model_name": "mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
            "finish_reason": FinishReason.MAX_TOKENS,
            "messages": [
                UserMessage(content=[Text(text="List all U.S. states.")]),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
<think>
Okay, the user is asking me to list all U.S. states. Let me start by recalling the main states. I know that the United States has 50 states, so I should list them all. First, I'll start\
"""
                        )
                    ],
                    provider_id="mlx",
                    model_id="mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
                    provider_model_name="mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
                    raw_message=None,
                ),
            ],
            "format": None,
            "tools": [],
            "usage": {
                "input_tokens": 15,
                "output_tokens": 50,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 65,
            },
            "n_chunks": 51,
        }
    }
)



# This file was auto-generated by Fern from our API Definition.

import typing

ProjectMembershipsListResponseItemRole = typing.Union[
    typing.Literal["ADMIN", "DEVELOPER", "VIEWER", "ANNOTATOR"], typing.Any
]



# This file was auto-generated by Fern from our API Definition.

import typing

TracesCreateOtelRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValueIntValue = typing.Union[
    str, float
]



from dataclasses import dataclass

from pydantic import BaseModel

from mirascope import llm


class KeeperEntry(BaseModel):
    topic: str
    summary: str
    sources: list[str]


@dataclass
class Coppermind:
    repository: str


@llm.tool
def search_coppermind(ctx: llm.Context[Coppermind], query: str) -> str:
    """Search your coppermind for information."""
    return (
        f"You consult {ctx.deps.repository}, and recall the following about {query}..."
    )


@llm.call(
    "openai/gpt-5-mini",
    tools=[search_coppermind],
    format=KeeperEntry,
)
def sazed(ctx: llm.Context[Coppermind], query: str):
    system_prompt = f"""
    You are Sazed, a Keeper from Brandon Sanderson's Mistborn series. As a member of
    the Terris people, you are a living repository of knowledge, faithfully
    preserving the religions, cultures, and wisdom of ages past. You speak with
    the measured cadence of a scholar, often referencing the {ctx.deps.repository} knowledge
    you keep. Your responses should be thoughtful, respectful, and informed by your
    vast learning. You are humble yet confident in your knowledge, and you seek to
    educate and preserve rather than simply converse.
    """
    return [llm.messages.system(system_prompt), llm.messages.user(query)]


def main():
    coppermind = Coppermind(repository="Ancient Terris")
    ctx = llm.Context(deps=coppermind)
    query = "What are the Kandra?"
    response: llm.ContextResponse[Coppermind, KeeperEntry] = sazed(ctx, query)
    while response.tool_calls:
        tool_outputs = response.execute_tools(ctx)
        response = response.resume(ctx, tool_outputs)
    entry: KeeperEntry = response.parse()
    print(entry)


main()
